// program ct_sino
// 
// Required files are q???.img, dark.img.
// output file "s????.sin"

/*----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <time.h>

/*----------------------------------------------------------------------*/
#ifndef M_PI
#define M_PI				3.1415926535897932385
#endif
#define INTEL
#define HiPic_Header_Size	64
#define MAX_SHOT			9000
#define MAXPIXL				5001

/*----------------------------------------------------------------------*/
struct HiPic_Header{
	char			head[2];
	short			comment_length;
	short			width;
	short			height;
	short			x_offset;
	short			y_offset;
	short			type;
	char			reserved[63 - 13];
	char			*comment;
};
typedef struct HiPic_Header Header;

/*----------------------------------------------------------------------*/

// main data for read transmitted images (data[y][x])
short	data[MAXPIXL];

//dark file name and data
short	dark[MAXPIXL];
short	cent_flag;

// image profile from 'output.log'
float	shottime[MAX_SHOT], shotangle[MAX_SHOT];
int		Ishot[MAX_SHOT], NST, NI0, II0[MAX_SHOT];

// parameters for interpolation of incident beam
short	II01[MAXPIXL], II02[MAXPIXL];     // II01[x], II02[x]
short	I[MAXPIXL];                                // I[x]
double	I0[MAXPIXL];

// file head character  q???.img or q????.img
char	flhead[]   = "q";
char	darkfile[] = "dark.img";

// output file name
char	fn[15];

// flag for q001.img or q0001.img
int		iFlag;

// projection
short	N, n_total;

// calculation layer
long	s_layer, b_layer;

// total absorption
double	TA, TA1, TA2, TAM, TASD;

/*----------------------------------------------------------------------*/

static void Error(msg)
char        *msg;
{
	fputs(msg,stderr);
	fputc('\n',stderr);
	exit(1);
}


/*----------------------------------------------------------------------*/
int read_hipic(char *fname, short *data, Header *h, long ln)
{
	int		j, i_res;
	FILE	*fi;

//open input files
	if((fi = fopen(fname,"rb")) == NULL){
		printf("can not open %s for input\n", fname);
		free(h->comment); 
		return(-10);
	}
// read comment and image data from file1
	if (fread(h, sizeof(char), HiPic_Header_Size, fi) != HiPic_Header_Size){
		fprintf(stderr, "EOF in %s (header)\n", fname);
		free(h->comment); 
		return(-1);
	}
	if (strncmp(h->head, "IM", 2)){
		fclose(fi); 
		free(h->comment); 
		fprintf(stderr, "This is not HiPIc image file.\n"); 
		fprintf(stderr, "File name = %s\n", fname); 
		return(-11);
	}

	h->comment = (char *) malloc(h->comment_length + 1);
	if (h->comment == NULL){
		fclose(fi); 
		free(h->comment); 
		fprintf(stderr, "Connot allocate memory (HiPic_Header.comment)\n"); 
		fprintf(stderr, "File name = %s\n", fname); 
		return(-12);
	}
	if (fread(h->comment, sizeof(char), h->comment_length, fi) != (unsigned short)h->comment_length){ 
		fclose(fi); 
		free(h->comment); 
		fprintf(stderr, "EOF found (comment)\n"); 
		fprintf(stderr, "File name = %s\n", fname); 
		return(-13);
	}
//	x_size = h->width;
//	y_size = h->height;
	h->comment[h->comment_length] = '\0';
	i_res = fseek( fi, h->width*ln*sizeof(short), SEEK_CUR);
	if( i_res != 0 ){
		fclose(fi);
		free(h->comment); 
		fprintf(stderr, " Error seeking %s.\n", fname);
		fprintf(stderr, "File name = %s\n", fname); 
		return(ln+1);
	}
	else{
		if ((j = fread(data, sizeof(short), h->width, fi)) != h->width){
			fclose(fi);
			free(h->comment); 
			fprintf(stderr, " Error reading %s at %d (%d)\n", fname, ln, j);
			fprintf(stderr, "File name = %s\n", fname); 
			return(ln+1);
		}
	}

// check for correct reading of HiPic file
//printf("-------------------------------\n");
//printf("%s\n", fname);
//printf("%s\n", h->head);
//printf("%d\n", h->comment_length);
//printf("%d\n", h->width);
//printf("%d\n", h->height);
//printf("%d\n", h->x_offset);
//printf("%d\n", h->y_offset);
//printf("%d\n", h->type);
//printf("%s\n", h->reserved);
//printf("%s\n", h->comment);
//printf("-------------------------------\n");
	N = h->width;
	free(h->comment);
	fclose(fi);
	return(0);
}

/*----------------------------------------------------------------------*/

int read_log()
{
	int			i, j;
	short		nnn;
	FILE		*f;
	char		line[100], *ss;
	int			flg_I0;
// open parameter file
	if((f = fopen("output.log","r")) == NULL){
		printf("cannot open output.log \n");
		return(-10);
	}

// read parameters
	NI0 = 0;
	NST = 0;
	nnn = 0;
	j   = 0;
	while(j>=0){
		j=j+1;
		line[0]='\0';
		ss = fgets(line, 100, f);
		if (i = strlen(line) > 2){
			sscanf(line, "%d %f %f %d", &nnn, &shottime[j], &shotangle[j], &flg_I0);
			if (flg_I0 == 0){
				II0[NI0] = nnn;
				NI0 = NI0 + 1;				// number of I_0 (direct beam)
			}
			else{
				Ishot[NST] = nnn;
				NST = NST + 1;				// number of I (sample image)
			}
		}
		else{
			break;
		}
	}
	n_total = (short)(nnn + 2);
	fprintf(stderr, " nshot = %d, NI0 = %d, total = %d \n", NST, NI0, n_total);
	fclose(f);
	if(n_total<1000) iFlag = 0;
	if(n_total>=1000) iFlag = 1;
	if(n_total>MAX_SHOT){
		fprintf(stderr, "Too many projections!");
		return(1);
	}
	return(0);
}

/*----------------------------------------------------------------------*/

static int StoreProjection(ln)
long		ln;
{
	int			i, j, k, jx, nshot, ilp;
	double		t1, t2;
	double		a[MAXPIXL], b[MAXPIXL];
	Header		h;
	char		ffi01[15], ffi02[15], fii[15];
	double		I01, I02;

	double		*po;
	double		p_sum,p_ave;
	FILE		*fi;

//open output files
	sprintf(fn,"s%04d.sin", ln);
	if((fi = fopen(fn,"wb")) == NULL){ 
		printf("can not open %s for input\n", fn);
		return(-10);
	}

// p ̏
	po = (double *)malloc(N * sizeof(double));
	TA1 =0.0;
	TA2 =0.0;

// ߑ̐JEgB(ŏIIɂ NST = nshot ɂȂ͂)
	nshot = 0;

// I_0  I_0 ɋ܂ꂽߑ荞ރ[v
	for ( j = 0; j < NI0-1; ++j){

//IIO[j] and IIO[j+1]  open 
		if(iFlag==0) sprintf(ffi01, "%s%03d.img", flhead, II0[j]);
		if(iFlag==1) sprintf(ffi01, "%s%04d.img", flhead, II0[j]);
		if ((i = read_hipic(ffi01, &II01[0], &h, ln)) != 0){
			printf("something wrong -- return value is %d(II01)", i);
			exit(-1);
		}
		if(iFlag==0) sprintf(ffi02, "%s%03d.img", flhead, II0[j+1]);
		if(iFlag==1) sprintf(ffi02, "%s%04d.img", flhead, II0[j+1]);
		if ((i = read_hipic(ffi02, &II02[0], &h, ln)) != 0){
			printf("something wrong -- return value is %d(II02)", i);
			exit(-1);
		}
// I0EV ̌W߂邽߂̏
		t1 = shottime[II0[j]];
		t2 = shottime[II0[j+1]];

// 1 layer(ln) I_0̒l}
		for (jx = 0; jx < N; ++jx){
			I01   = (double)II01[jx]-dark[jx]; // OI_0ňꎟ
			I02   = (double)II02[jx]-dark[jx];
			a[jx] = (double)(((double)(I02   - I01))    / (t2 - t1));      //ԌW߂
			b[jx] = (double)(((double)I01*t2 - (double)I02*t1) / (t2 - t1));
		}
		for ( k = II0[j] + 1; k < II0[j+1]; ++k){
// a[jx], b[jx]  shottime[k] p p(x) ߂B
			if(iFlag==0) sprintf(fii, "%s%03d.img", flhead, k);
			if(iFlag==1) sprintf(fii, "%s%04d.img", flhead, k);
			if ((i = read_hipic(fii, &I[0], &h, ln)) != 0){
				printf("something wrong -- return value is %d(II01)", i);
				exit(-1);
			}
			ilp=0;
			for (jx = 0; jx < N; ++jx){
				I0[jx]    = a[jx] * shottime[k] + b[jx];
				if ((I[jx]-dark[jx]) < 1){
					if(ilp==0){
						printf("Warning \t");
//						printf("  jx = %d, I0 = %f, I = %d, dark = %d, ln =%d \n", jx, I0[jx], I[jx], dark[jx], ln);
//						printf("  II01 = %d, II02 = %d\n", II01[jx], II02[jx]);
//						printf("  t1   = %f, t2   = %f \n", t1, t2);
//						printf("  data is negative!! = %f,  %d\n",(double)I0[jx] / (double)(I[jx]-dark[jx]), jx);
//						printf("  a = %f,  b = %f\n", a[jx], b[jx]);
						printf("  %s\tare are\n", fii);
						ilp=1;
					}
				}
				*(po+jx)= log((double)(I0[jx])/(double)(I[jx]-dark[jx]));
			}
			if(nshot<b_layer) ilp=1;
			if(ilp==1){
				for(jx=0;jx<N;++jx){
					*(po+jx)=0.0;
				}
			}

// roupe for correction
			p_sum = 0.0;
			p_ave = 0.0;
			for (jx=0; jx<10; ++jx){
				p_sum = p_sum + *(po+jx) + *(po+N-jx);
			}
			p_ave = p_sum / (10.-0.) / 2.;
			TA=0.0;
			for (jx=0; jx<N; ++jx){
				*(po+jx) = *(po+jx) - p_ave; /* comment out here for Local CT*/
				TA       = TA + *(po+jx);
			}
			TA1=TA1+TA;
			TA2=TA2+TA*TA;
//
			if (i = fwrite(po , sizeof(double), N, fi) > N*sizeof(double)){
				fprintf(stderr, " Error writing data to %s at %d\n", fn, i);
				fclose(fi);
				return(i+1);
			}
			nshot = nshot + 1;
		} // end of k loop
	} // end of j loop
	fclose(fi);
	return (0);
}

float calc_center(){
int		i;
long	j;
double	mm1, mx1, mm2, mx2;
double	*p000, *p180, rsum, rd;
float	center;
Header	h;

// file name and data for calculation of rotation center (data[y][x])
// odd = I_0, even = transmitted
char	c_file[13];
short	cent_1[MAXPIXL], cent_2[MAXPIXL];
short	cent_3[MAXPIXL], cent_4[MAXPIXL];

// read center file
//I_0 file
	if(iFlag==0) sprintf(c_file, "%s001.img", flhead);
	if(iFlag==1) sprintf(c_file, "%s0001.img", flhead);
	if ((i = read_hipic(c_file, &cent_1[0], &h, 0)) != 0){
		printf("something wrong -- return value is %d", i);
		exit(-1);
	}

// zero degree file(transmitted)
	if(iFlag==0) sprintf(c_file, "%s002.img", flhead);
	if(iFlag==1) sprintf(c_file, "%s0002.img", flhead);
	if ((i = read_hipic(c_file, &cent_2[0], &h, 0)) != 0){
		printf("something wrong -- return value is %d", i);
		exit(-1);
	}

//I_0 file
	if(iFlag==0) sprintf(c_file, "%s%03d.img", flhead, n_total-1);
	if(iFlag==1) sprintf(c_file, "%s%04d.img", flhead, n_total-1);
	if ((i = read_hipic(c_file, &cent_3[0], &h, 0)) != 0){
		printf("something wrong -- return value is %d", i);
		exit(-1);
	}

//  180 degree file(transmitted)
	if(iFlag==0) sprintf(c_file, "%s%03d.img", flhead, n_total);
	if(iFlag==1) sprintf(c_file, "%s%04d.img", flhead, n_total);
	if ((i = read_hipic(c_file, &cent_4[0], &h, 0)) != 0){
		printf("something wrong -- return value is %d", i);
		exit(-1);
	}


	center = 0.0;
//calculate gravity center
		mm1 = 0.0;
		mx1 = 0.0;
		mm2 = 0.0;
		mx2 = 0.0;
		for (j=0; j<h.width; ++j){
			mx1 = mx1 + (log((double)(cent_2[j]-dark[j]) / (double)(cent_1[j]-dark[j]))) * (double)j;
			mm1 = mm1 +  log((double)(cent_2[j]-dark[j]) / (double)(cent_1[j]-dark[j]));
			mx2 = mx2 + (log((double)(cent_4[j]-dark[j]) / (double)(cent_3[j]-dark[j]))) * (double)j;
			mm2 = mm2 +  log((double)(cent_4[j]-dark[j]) / (double)(cent_3[j]-dark[j]));
		}
		center = (float)((mx1/mm1+ mx2/mm2)/2.);
	//printf("%d %f %f %f %f\n", j, mm1, mx1, mm2, mx2);

// calculate center with RMSD
	rsum=0.0;
	rd=0.0;
	p000 = (double *)malloc(N * sizeof(double));
	p180 = (double *)malloc(N * sizeof(double));
	for (j=0; j<N; ++j){
		*(p000+j)=log((double)(cent_2[j]-dark[j]) / (double)(cent_1[j]-dark[j]));
		*(p180+j)=log((double)(cent_4[j]-dark[j]) / (double)(cent_3[j]-dark[j]));
	}

	
	return(center);
}


/*----------------------------------------------------------------------*/

#ifndef CLOCKS_PER_SEC
#define CLOCKS_PER_SEC	1000000
extern long clock();
#endif

#define CLOCK()		((double)clock()/(double)CLOCKS_PER_SEC)
#define DESC_LEN	256

int main(argc,argv)
int		argc;
char	**argv;
{
	double		Clock;					// timer setting
	int			i;
	Header		h;
	FILE		*fo;

// parameter setting
	if (argc==3){
	}
	else{
		fprintf(stderr, "parameter was wrong!!!\n");
		fprintf(stderr, "usage : %s layer blackarea\n", argv[0]);
		exit(1);
	}
	s_layer = atoi(argv[1]);
	b_layer = atoi(argv[2]);

// read shot log file
	if ((i = read_log()) != 0){
		fprintf(stderr, "canot read 'output.log' ! (%d)\n", i);
		exit(1);
	}

// read dark image
	if ((i = read_hipic(darkfile, &dark[0], &h, s_layer)) != 0){
		fprintf(stderr, "something wrong in dark file (%d)\n", i);
		exit(1);
	}
//	fprintf(stderr, "width of projection data = %d \n",  h.width);

	cent_flag = 0;
	Clock=CLOCK();
	if((i=StoreProjection(s_layer)) !=0){
		printf("something wrong in StoreProjection (%d)\n",i);
		exit(1);
	}
	TAM=TA1/NST;
	TASD=sqrt(TA2/NST-TAM*TAM);
	fprintf(stderr," TA mean\t%lf\tTASD\t%lf\n", TAM, TASD);
	fprintf(stderr," TASD/TA mean\t%lf\n", TASD/TAM);
	fprintf(stderr," Store Sinogram \t%lf / sec\n",CLOCK()-Clock);
//	fprintf(stderr, "output file name = %s \n", fn);

//open output files
	if((fo = fopen("sino.tmp","w")) == NULL){ 
		printf("can not open %s for input\n", fn);
		exit(1);
	}
	fprintf(fo, "%d\n", h.width);
	fprintf(fo, "%d\n", NST);
	fprintf(fo, "%04d\n", s_layer);
//	fprintf(fo, "%f\n", calc_center());
	fprintf(fo, "%f\n", h.width/2.);
	fprintf(fo, "%lf\n", TAM); // Mean of Total Absorption from sinogram
	fprintf(fo, "%lf\n", TASD); // SD of Total Absorption from sinogram
	fclose(fo);

	return 0;
}
