// program ct_sino
// 
// Required files are q???.img, dark.img.
// output file "s????.sin"

/*----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include <malloc.h>

/*----------------------------------------------------------------------*/
#ifndef M_PI
#define M_PI				3.1415926535897932385
#endif
#define INTEL
#define HiPic_Header_Size	64
#define MAX_SHOT			9000
#define MAXPIXL				5001
#define MAXLINE				2000

/*----------------------------------------------------------------------*/
struct HiPic_Header{
	char			head[2];
	short			comment_length;
	short			width;
	short			height;
	short			x_offset;
	short			y_offset;
	short			type;
	char			reserved[63 - 13];
	char			*comment;
};
typedef struct HiPic_Header Header;

/*----------------------------------------------------------------------*/

// main data for read transmitted images (data[y][x])
short	*data;

//dark file name and data
short	*dark;

// image profile from 'output.log'
float	shottime[MAX_SHOT], shotangle[MAX_SHOT], numl;
short	Ishot[MAX_SHOT], NST, NI0, II0[MAX_SHOT], flg_I0[MAX_SHOT];

// parameters for interpolation of incident beam
//short	*II01, *II02;     // II01[x], II02[x]
//short	*I;                                // I[x]

// file head character  q???.img or q????.img
char	flhead[]   = "q";
char	darkfile[] = "dark.img";

// output file name
char	fn[15];

// flag for q001.img or q0001.img
int		iFlag;

// projection
short	N, n_total, x_size, y_size;

// calculation layer
long	s_layer;


/*----------------------------------------------------------------------*/

static void Error(msg)
char        *msg;
{
	fputs(msg,stderr);
	fputc('\n',stderr);
	exit(1);
}


/*----------------------------------------------------------------------*/


int read_head(char *fname, Header *h)
{
	int		j, i_res;
	FILE	*fi;

//open input files
	if((fi = fopen(fname,"rb")) == NULL){
		printf("can not open %s for input\n", fname);
		free(h->comment); 
		return(-10);
	}
// read comment and image data from file1
	if(fread(h, sizeof(char), HiPic_Header_Size, fi) != HiPic_Header_Size){
		fprintf(stderr, "EOF in %s (header)\n", fname);
		free(h->comment); 
		return(-1);
	}

	fclose(fi);
	return(0);
}

/*----------------------------------------------------------------------*/
int read_hipic(char *fname, Header *h)
{
	int		j, i_res;
	FILE	*fi;

//open input files
	if((fi = fopen(fname,"rb")) == NULL){
		printf("can not open %s for input\n", fname);
		free(h->comment); 
		return(-10);
	}
// read comment and image data from file1
	if (fread(h, sizeof(char), HiPic_Header_Size, fi) != HiPic_Header_Size){
		fprintf(stderr, "EOF in %s (header)\n", fname);
		free(h->comment); 
		return(-1);
	}
	if (strncmp(h->head, "IM", 2)){
		fclose(fi); 
		free(h->comment); 
		fprintf(stderr, "This is not HiPIc image file.\n"); 
		fprintf(stderr, "File name = %s\n", fname); 
		return(-11);
	}

	h->comment = (char *) malloc(h->comment_length + 1);
	if (h->comment == NULL){
		fclose(fi); 
		free(h->comment); 
		fprintf(stderr, "Connot allocate memory (HiPic_Header.comment)\n"); 
		fprintf(stderr, "File name = %s\n", fname); 
		return(-12);
	}
	if (fread(h->comment, sizeof(char), h->comment_length, fi) != (unsigned short)h->comment_length){ 
		fclose(fi); 
		free(h->comment); 
		fprintf(stderr, "EOF found (comment)\n"); 
		fprintf(stderr, "File name = %s\n", fname); 
		return(-13);
	}
	x_size = h->width;
	y_size = h->height;
	h->comment[h->comment_length] = '\0';
	i_res = fseek( fi, h->width*s_layer*sizeof(short), SEEK_CUR);
	if( i_res != 0 ){
		fclose(fi);
		free(h->comment); 
		fprintf(stderr, " Error seeking %s.\n", fname);
		fprintf(stderr, "File name = %s\n", fname); 
		return(s_layer+1);
	}
	else{
		if ((j = fread(data, sizeof(short), h->width, fi)) != h->width){
			fclose(fi);
			free(h->comment); 
			fprintf(stderr, " Error reading %s at %d\n", fname, j);
			fprintf(stderr, "File name = %s\n", fname); 
			return(s_layer+1);
		}
	}

// check for correct reading of HiPic file
//printf("-------------------------------\n");
//printf("%s\n", fname);
//printf("%s\n", h->head);
//printf("%d\n", h->comment_length);
//printf("%d\n", h->width);
//printf("%d\n", h->height);
//printf("%d\n", h->x_offset);
//printf("%d\n", h->y_offset);
//printf("%d\n", h->type);
//printf("%s\n", h->reserved);
//printf("%s\n", h->comment);
//printf("-------------------------------\n");
	N = h->width;
	free(h->comment);
	fclose(fi);
	return(0);
}

/*----------------------------------------------------------------------*/

int read_log()
{
	int			i, j;
	short		nnn;
	FILE		*f;
	char		line[100], *ss;
// open parameter file
	if((f = fopen("output.log","r")) == NULL){
		printf("cannot open output.log \n");
		return(-10);
	}

// read parameters
	NI0 = 0;
	NST = 0;
	nnn = 0;
	j   = 0;
	while(j>=0){
		j=j+1;
		line[0]='\0';
		ss = fgets(line, 100, f);
		if (i = strlen(line) > 10){
			sscanf(line, "%d %f %f %d", &nnn, &shottime[j], &shotangle[j], &flg_I0[j]);
			if (flg_I0[j] == 0){
				II0[NI0] = nnn;
				NI0 = NI0 + 1;				// I_0  shot 
			}
			else{
				Ishot[NST] = nnn;
				NST = NST + 1;				// I  shot 
			}
		}
		else{
			break;
		}
	}
	n_total = (short)(nnn);
	fprintf(stderr, " nshot = %d, NI0 = %d, total = %d \n", NST, NI0, n_total);
	fclose(f);
	if(n_total+2<1000) iFlag = 0;
	if(n_total+2>=1000) iFlag = 1;
	if(n_total+2>MAX_SHOT){
		fprintf(stderr, "Too many projections!");
		return(1);
	}
	return(0);
}

/*----------------------------------------------------------------------*/

static int StoreProjection()
{
	int			i, j, k, jx, m, nshot, ilp;
	Header		h;
	char		ffi[15];

	unsigned short		*po;
	FILE		*fo[MAXLINE];

// read dark image
	data=(short *)malloc(x_size*numl*sizeof(short));
	if ((i = read_hipic(darkfile, &h)) != 0){
		fprintf(stderr, "something wrong in dark file (%d)\n", i);
		exit(1);
	}
	dark=(short *)malloc(x_size*numl*sizeof(short));
	for(i=0;i<x_size*numl;++i) *(dark+i)=*(data+i);
	free(data);

//open output files
	for(m=0;m<numl;++m){
		sprintf(fn,"s%04d.sin", (int)(s_layer+m));
		if((fo[m] = fopen(fn,"wb")) == NULL){ 
			printf("can not open %s for output\n", fn);
			return(-10);
		}

		if (i = fwrite(&N , sizeof(short), 1, fo[m]) > sizeof(short)){
			fprintf(stderr, " Error writing N to %s at %d\n", fn, i);
			fclose(fo[m]);
			return(1);
		}
		if (i = fwrite(&n_total , sizeof(short), 1, fo[m]) > sizeof(short)){
			fprintf(stderr, " Error writing n_total to %s at %d\n", fn, i);
			fclose(fo[m]);
			return(1);
		}
	}


// p ̏
	po = (unsigned short *)malloc(N * sizeof(unsigned short));

// Xvbg郋[v
	for (j=1;j<n_total+1;++j){

		if(iFlag==0) sprintf(ffi, "%s%03d.img", flhead, j);
		if(iFlag==1) sprintf(ffi, "%s%04d.img", flhead, j);
		data=(short *)malloc(x_size*numl*sizeof(short));
		if ((i = read_hipic(ffi, &h)) != 0){
			printf("something wrong -- return value is %d(II01)", i);
			return(1);
		}

		for(m=0;m<numl;++m){
			if (i = fwrite(&flg_I0[j], sizeof(short), 1, fo[m]) > sizeof(short)){
				fprintf(stderr, " Error writing data to %s at %d\n", fn, i);
				fclose(fo[m]);
				return(i+1);
			}
			if (i = fwrite(&shottime[j], sizeof(float), 1, fo[m]) > sizeof(float)){
				fprintf(stderr, " Error writing data to %s at %d\n", fn, i);
				fclose(fo[m]);
				return(i+1);
			}
			if (i = fwrite(&shotangle[j], sizeof(float), 1, fo[m]) > sizeof(float)){
				fprintf(stderr, " Error writing data to %s at %d\n", fn, i);
				fclose(fo[m]);
				return(i+1);
			}
			
			for(jx=0;jx<N;++jx){
				if((*(data+N*m+jx)-*(dark+N*m+jx))>0){
					*(po+jx)=*(data+N*m+jx)-*(dark+N*m+jx);
				}else{
					*(po+jx)=0;
				}
			}
			if (i = fwrite(po, sizeof(unsigned short), N,  fo[m]) > N*sizeof(unsigned short)){
				fprintf(stderr, " Error writing data to %d at %d\n", j, m);
				fclose(fo[m]);
				return(i+1);
			}
//				printf("%d\t%d\t%d\t%d\n", j,m,jx,*(po+jx));
			printf("%d\t%d\t%d\t%d\n", m,*(po+jx+10),*(po+jx+500),*(po+jx+1000));
			
		}
		free(data);
//		printf("%d\t%d\t%d\n",j,*(po),*(po+100));

	} // end of j loop
	for(m=0;m<numl;++m){
		fclose(fo[m]);
	}
	free(dark);
	free(po);
	return (0);
}


/*----------------------------------------------------------------------*/

#ifndef CLOCKS_PER_SEC
#define CLOCKS_PER_SEC	1000000
extern long clock();
#endif

#define CLOCK()		((double)clock()/(double)CLOCKS_PER_SEC)
#define DESC_LEN	256

int main(argc,argv)
int		argc;
char	**argv;
{
	double		Clock;					// timer setting
	int			i;
	Header		h;
	FILE		*fo;

// parameter setting
	if (argc==3){
		numl=atoi(argv[2]);
		if(numl<0) numl=1;
		if(numl>MAXLINE) numl=MAXLINE;
	}
	else if (argc==2){
		numl=1;
	}
	else{
		fprintf(stderr, "usage : %s layer num\n", argv[0]);
		return(1);
	}
	s_layer = atoi(argv[1]);

	sprintf(flhead,   "%s", "q");
	sprintf(darkfile, "%s", "dark.img");

// check layer number
	if((i=read_head(darkfile, &h)) != 0){
		fprintf(stderr, "error check layer (%d)\n", i);
		exit(1);
	}
	x_size=h.width; y_size=h.height;
	if(y_size<s_layer){
		Error("Too large layer number");
	}
	if(y_size<s_layer+numl){
		numl=y_size-s_layer;
		fprintf(stderr, "number of stack was changed to %d\n", (int)numl);
	}

// read shot log file
	if ((i = read_log()) != 0){
		fprintf(stderr, "canot read 'output.log' ! (%d)\n", i);
		return(1);
	}

//	fprintf(stderr, "width of projection data = %d \n",  h.width);

	Clock=CLOCK();
	if((i=StoreProjection()) !=0){
		printf("something wrong in StoreProjection (%d)\n",i);
		return(1);
	}
	fprintf(stderr," Store Sinogram \t%lf / sec\n",CLOCK()-Clock);
//	fprintf(stderr, "output file name = %s \n", fn);

	return 0;
}
